Imports vgm2mid.clsMID
Public Class clsSN76489
    ' Base oscillator frequency
    Private Const OSC1 = 53693100

    ' Sample frequency (fsam)
    Public Const fM_PSG = OSC1 / 15
    Public Const fsam_PSG = fM_PSG / 72

    ' Command constants
    Public Const PSG_CHANNEL_SELECT = &HF0 '240
    Public Const PSG_TONE_1 = &H80 '128
    Public Const PSG_TONE_2 = &HA0 '160
    Public Const PSG_TONE_3 = &HC0 '192
    Public Const PSG_NOISE = &HE0 '224
    Public Const PSG_ATTENUATOR = &H70 '112

    Public Const PSG_FB_PERIODIC = &H0
    Public Const PSG_FB_WHITE = &H1

    Public Const PSG_CLOCK_SOURCE_HALF = &H0
    Public Const PSG_CLOCK_SOURCE_FOURTH = &H1
    Public Const PSG_CLOCK_SOURCE_EIGHTH = &H2
    Public Const PSG_CLOCK_SOURCE_TONE_3 = &H3

    Public Const PSG_ATTENUATOR_1 = &H90 '144
    Public Const PSG_ATTENUATOR_2 = &HB0 '176
    Public Const PSG_ATTENUATOR_3 = &HD0 '208
    Public Const PSG_ATTENUATOR_NOISE = &HF0 '240
    Public Const PSG_ATTENUATION_MIN = &H0
    Public Const PSG_ATTENUATION_MAX = &HF

    Public Const MIDI_CHANNEL_PSG_BASE = 10

    Public Sub Init(ByRef MIDI As clsMID)
        Dim ProgramChange As New clsMID.Commands.ProgramChange()
        Dim i As Integer
        For i = 10 To 12
            With ProgramChange
                .Channel = i : .Program = Programs.Lead1Square
            End With
            MIDI.AddEvent(ProgramChange)
        Next
    End Sub

    Private Function toHz(ByVal FNum As Double, Optional ByVal Block As Byte = 0) As Double
        If FNum = 0 Then Return 0 Else Return ((fM_PSG / 32) / FNum)
    End Function

    Public Sub ProcessCommand(ByVal MSB As Byte, ByVal LSB As Byte, ByRef MIDI As clsMID)
        Dim CH As Byte, MIDIChannel As Byte
        Static FNum_1(3) As Long, FNum_2(3) As Long, Hz_1(3) As Double, Hz_2(3) As Double, Note_1(3) As Double, Note_2(3) As Double
        Static FB As Byte, ClockSource As Byte
        Static Attenuation_1(3) As Byte, Attenuation_2(3) As Byte
        Static MIDINote(3) As Byte, MIDIWheel(3) As Integer, NoteOn_1(3) As Byte, NoteOn_2(3) As Byte, MIDIVolume As Byte

        Static NoteOff As New Commands.NoteOff()
        Static NoteOn As New Commands.NoteOn()
        Static ControllerChange As New Commands.ControllerChange()

        Select Case (MSB And PSG_CHANNEL_SELECT)
            Case PSG_TONE_1, PSG_TONE_2, PSG_TONE_3
                CH = Fix((MSB And &H70) / &H20)

                'If PSG_CH_DISABLED(CH) = 1 Then Exit Sub

                'CH = IIf(CH = 2 And ClockSource = PSG_CLOCK_SOURCE_TONE_3, 3, CH)

                MIDIChannel = MIDI_CHANNEL_PSG_BASE + CH 'IIf(CH = 2 And ClockSource = PSG_CLOCK_SOURCE_TONE_3, 3, CH)

                FNum_1(CH) = FNum_2(CH)
                FNum_2(CH) = ((LSB And &H3F) * &H10) + (MSB And &HF)

                Hz_1(CH) = Hz_2(CH)
                Hz_2(CH) = toHz(FNum_2(CH))

                Note_1(CH) = Note_2(CH)
                Note_2(CH) = MIDI.toNote(Hz_2(CH))

                If FNum_2(CH) = 0 Then
                    If NoteOn_2(CH) <> 0 Then
                        With NoteOff
                            .Channel = MIDIChannel : .Note = MIDINote(CH) : .Velocity = 0
                        End With
                        MIDI.AddEvent(NoteOff)
                        With ControllerChange
                            .Channel = MIDIChannel : .Controller = Controllers.Volume : .ControllerValue = 0
                        End With
                        MIDI.AddEvent(ControllerChange)
                        NoteOn_1(CH) = NoteOn_2(CH)
                        NoteOn_2(CH) = 0
                    End If
                ElseIf FNum_2(CH) > 0 And FNum_2(CH) <> FNum_1(CH) Then
                    If NoteOn_2(CH) = 0 Then
                        MIDI.ChangePitch(Note_1(CH), Note_2(CH), MIDIChannel, MIDINote(CH), MIDIWheel(CH), True)
                    Else
                        MIDI.ChangePitch(Note_1(CH), Note_2(CH), MIDIChannel, MIDINote(CH), MIDIWheel(CH))
                    End If
                    NoteOn_1(CH) = NoteOn_2(CH)
                    NoteOn_2(CH) = 1
                End If

            Case PSG_NOISE
                CH = 3

                MIDIChannel = MIDI_CHANNEL_PSG_BASE + CH

                FB = (MSB And 4) / 4
                ClockSource = MSB And 3

                'If ClockSource <> PSG_CLOCK_SOURCE_TONE_3 Then
                '    MIDIChannel = MIDI_CHANNEL_PSG_BASE + 2
                '    Dim intNote As Integer
                '    For intNote = 0 To 127
                '        MIDI_AddEvent MIDI_NOTE_OFF, MIDIChannel, intNote, &H0
                '    Next
                'End If

            Case PSG_ATTENUATOR_1, PSG_ATTENUATOR_2, PSG_ATTENUATOR_3 ', PSG_ATTENUATOR_NOISE
                CH = (((MSB And PSG_ATTENUATOR) / &H10) - 1) / 2

                'If PSG_CH_DISABLED(CH) = 1 Then Exit Sub

                MIDIChannel = MIDI_CHANNEL_PSG_BASE + CH

                Attenuation_1(CH) = Attenuation_2(CH)
                Attenuation_2(CH) = (MSB And PSG_ATTENUATION_MAX) * 8.45 '(127 / 15)

                If Attenuation_2(CH) = PSG_ATTENUATION_MAX _
                And Attenuation_1(CH) <> PSG_ATTENUATION_MAX Then
                    With NoteOff
                        .Channel = MIDIChannel : .Note = MIDINote(CH) : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                    NoteOn_1(CH) = NoteOn_2(CH)
                    NoteOn_2(CH) = 0
                ElseIf Attenuation_1(CH) = PSG_ATTENUATION_MAX _
                And Attenuation_2(CH) <> PSG_ATTENUATION_MAX Then
                    MIDI.ChangePitch(Note_1(CH), Note_2(CH), MIDIChannel, MIDINote(CH), MIDIWheel(CH))
                    NoteOn_1(CH) = NoteOn_2(CH)
                    NoteOn_2(CH) = 1
                End If

                If Attenuation_2(CH) <> Attenuation_1(CH) Then
                    MIDIVolume = 127 - Attenuation_2(CH)
                    With ControllerChange
                        .Channel = MIDIChannel : .Controller = Controllers.Volume : .ControllerValue = MIDIVolume
                    End With
                    MIDI.AddEvent(ControllerChange)
                End If

            Case PSG_ATTENUATOR_NOISE
                'If PSG_NOISE_DISABLED = 1 Then Exit Sub

                CH = (((MSB And PSG_ATTENUATOR) / &H10) - 1) / 2

                MIDIChannel = 9 'CHN_DAC

                Attenuation_1(CH) = Attenuation_2(CH)
                Attenuation_2(CH) = (MSB And 15) * 8.45 '(127 / 15)
                MIDIVolume = 127 - Attenuation_2(CH)

                If MIDIVolume > 0 Then
                    With NoteOff
                        .Channel = MIDIChannel : .Note = Percussion.PedalHiHat : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                    With NoteOn
                        .Channel = MIDIChannel : .Note = Percussion.PedalHiHat : .Velocity = MIDIVolume
                    End With
                    MIDI.AddEvent(NoteOn)
                End If

                '    Case Else

        End Select

    End Sub

End Class