Imports System.IO
Public Class clsVGM

    'VGM v1.01 Specification
    '0x00: "Vgm " File identification
    '0x04: Relative offset to end of file (i.e. file length - 4)
    '0x08: Version number (current = 0x00000101)
    '0x0c: PSG chip speed in Hz (e.g. 3579545). 0 if there is no PSG chip used.
    '0x10: FM chip speed in Hz  (e.g. 3579545). 0 if there is no FM chip used.
    '0x14: Relative offset to GD3 tag or zero if no tag
    '0x18: Total numbers of samples (i.e. total of all wait values in the file)
    '0x1c: Relative offset to loop point, or 0 if no loop.
    '0x20: Number of samples in one loop, or 0 if there is no loop.
    '0x24: "Rate" of recording (usually 50, 60 or 0)
    '0x28-0x40 : padding (must be zero)

#Region " VGM Variables"

    Dim mbytVgm_(3) As Byte, mstrVgm_ As String, mintOffset As Integer, _
    mbytVersion(3) As Byte, mintHzPSG, mintHzFM, mintGD3Offset, mintTotalSamples, _
    mintLoopOffset, mintLoopSamples, mintRate As Integer

    Dim mstrFileName, mstrUnzippedFileName As String, mbytVGMData() As Byte
    Dim mstrConversionStatus As String

#End Region

#Region " VGM Properties "

    Public Property Vgm_() As String
        Get
            Return mstrVgm_
        End Get
        Set(ByVal Value As String)

        End Set
    End Property
    Public Property Offset() As Integer
        Get
            Return mintOffset
        End Get
        Set(ByVal Value As Integer)

        End Set
    End Property
    Public Property Version() As String
        Get
            Dim strVersion As String = _
            Str(mbytVersion(0)) & "." _
            & Str(mbytVersion(1)) & "." _
            & Str(mbytVersion(2)) & "." _
            & Str(mbytVersion(3))
            Return strVersion
        End Get
        Set(ByVal Value As String)

        End Set
    End Property
    Public Property HzPSG() As Integer
        Get
            Return mintHzPSG
        End Get
        Set(ByVal Value As Integer)

        End Set
    End Property
    Public Property HzFM() As Integer
        Get
            Return mintHzFM
        End Get
        Set(ByVal Value As Integer)

        End Set
    End Property
    Public ReadOnly Property GD3Offset() As Integer
        Get
            Return mintGD3Offset
        End Get
    End Property
    Public Property TotalSamples() As Integer
        Get
            Return mintTotalSamples
        End Get
        Set(ByVal Value As Integer)

        End Set
    End Property
    Public Property LoopOffset() As Integer
        Get
            Return mintLoopOffset
        End Get
        Set(ByVal Value As Integer)

        End Set
    End Property
    Public Property LoopSamples() As Integer
        Get
            Return mintLoopSamples
        End Get
        Set(ByVal Value As Integer)

        End Set
    End Property
    Public Property Rate() As Integer
        Get
            Return mintRate
        End Get
        Set(ByVal Value As Integer)

        End Set
    End Property
    Public Property FileName() As String
        Get
            Return mstrFileName
        End Get
        Set(ByVal Value As String)
            mstrFileName = Value
        End Set
    End Property
    Public ReadOnly Property ConversionStatus() As String
        Get
            Return mstrConversionStatus
        End Get
    End Property

#End Region

    'GD3 v1.00 Specification
    '0x00: "Gd3 " File identification
    '0x04: Four bytes: 0x00, 0x01, 0x00, 0x00
    '0x08: Total length of GD3 tag data
    '0x12-0x??: GD3 Data

#Region " GD3 Variables "

    Dim mbytGd3_(3) As Byte, mstrGd3_ As String, mintGD3Version, mintGD3Length As Integer, _
    mshtGD3Data() As Short, mstrGD3Data, _
    mstrTrackNameEnglish, mstrTrackNameJapanese, _
    mstrGameNameEnglish, mstrGameNameJapanese, _
    mstrSystemNameEnglish, mstrSystemNameJapanese, _
    mstrOriginalAuthorEnglish, mstrOriginalAuthorJapanese, _
    mstrGameReleaseDate, mstrVGMAuthor, mstrNotes As String

#End Region

#Region " GD3 Properties "

    Public Property GD3TrackNameEnglish()
        Get
            Return mstrTrackNameEnglish
        End Get
        Set(ByVal Value)

        End Set
    End Property
    Public Property GD3TrackNameJapanese()
        Get
            Return mstrTrackNameJapanese
        End Get
        Set(ByVal Value)

        End Set
    End Property
    Public Property GD3GameNameEnglish()
        Get
            Return mstrGameNameEnglish
        End Get
        Set(ByVal Value)

        End Set
    End Property
    Public Property GD3GameNameJapanese()
        Get
            Return mstrGameNameJapanese
        End Get
        Set(ByVal Value)

        End Set
    End Property
    Public Property GD3SystemNameEnglish()
        Get
            Return mstrSystemNameEnglish
        End Get
        Set(ByVal Value)

        End Set
    End Property
    Public Property GD3SystemNameJapanese()
        Get
            Return mstrSystemNameJapanese
        End Get
        Set(ByVal Value)

        End Set
    End Property
    Public Property GD3OriginalAuthorEnglish()
        Get
            Return mstrOriginalAuthorEnglish
        End Get
        Set(ByVal Value)

        End Set
    End Property
    Public Property GD3OriginalAuthorJapanese()
        Get
            Return mstrOriginalAuthorJapanese
        End Get
        Set(ByVal Value)

        End Set
    End Property
    Public Property GD3GameReleaseDate()
        Get
            Return mstrGameReleaseDate
        End Get
        Set(ByVal Value)

        End Set
    End Property
    Public Property GD3VGMAuthor()
        Get
            Return mstrVGMAuthor
        End Get
        Set(ByVal Value)

        End Set
    End Property
    Public Property GD3Notes()
        Get
            Return mstrNotes
        End Get
        Set(ByVal Value)

        End Set
    End Property

#End Region

#Region " GYM Variables "

    Dim mbytGYMData() As Byte

#End Region

    Protected Overrides Sub Finalize()
        mbytVGMData = Nothing
        mbytGYMData = Nothing
    End Sub

    Public Sub OpenFile(ByVal strFileName As String)
        mstrFileName = strFileName
        Dim strFileExt As String = LCase(Path.GetExtension(strFileName))
        Select Case strFileExt
            Case ".vgm", ".vgz"
                Try
                    If strFileExt = ".vgz" Then
                        UnzipFile(mstrFileName)
                    End If
                    GetHeaderData(mstrFileName)
                    GetGD3Data(mstrFileName)
                    GetVGMData(mstrFileName)
                    If strFileExt = ".vgz" Then
                        File.Delete(mstrFileName)
                        mstrFileName = strFileName
                    End If
                Catch ex As Exception
                    MsgBox(ex.Message)
                Finally
                    FileClose()
                End Try
            Case ".gym"
                GetGYMData(mstrFileName)
                ConvertGYMtoVGM()
            Case Else
        End Select
    End Sub

    Private Sub UnzipFile(ByRef strFileName As String)
        Dim strTempFileName As String = Path.GetTempPath & "temp.vgm.gz"
        File.Copy(strFileName, strTempFileName, True)
        Dim strAppPath As String = _
            Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly.Location)
        Dim strShellCmd As String = _
            "" & strAppPath & "\gzip.exe -d -f -q " & "" & strTempFileName & ""
        Shell(strShellCmd, AppWinStyle.MinimizedNoFocus, True)
        strTempFileName = Path.GetTempPath & "temp.vgm"
        Do Until Dir(strTempFileName) <> "" : Application.DoEvents() : Loop
        Do Until FileLen(strTempFileName) > 0 : Application.DoEvents() : Loop
        mstrUnzippedFileName = strTempFileName
        strFileName = mstrUnzippedFileName
        'mstrUnzippedFileName = Path.GetDirectoryName(strFileName) & "\" _
        '    & Path.GetFileNameWithoutExtension(strFileName) & ".vgm"
        'FileCopy(strTempFileName, mstrUnzippedFileName)
    End Sub

    Private Sub GetHeaderData(ByRef strFileName As String)
        'Verify file; if file is zipped, file will be unzipped
        Dim strFileExt As String = LCase(Path.GetExtension(strFileName))
        If (strFileExt = ".vgz") And (mstrUnzippedFileName = "") Then
            UnzipFile(strFileName)
        End If

        'Read file header
        FileOpen(1, strFileName, OpenMode.Binary)
        FileGet(1, mbytVgm_)

        'Verify "Vgm "
        Dim strVgm_ As String, i As Integer
        For i = 0 To 3
            strVgm_ &= Chr(mbytVgm_(i))
        Next
        If Not (strVgm_ = "Vgm ") Then
            Exit Sub
        End If
        mstrVgm_ = strVgm_
        FileGet(1, mintOffset)
        FileGet(1, mbytVersion)
        FileGet(1, mintHzPSG)
        FileGet(1, mintHzFM)
        FileGet(1, mintGD3Offset)
        FileGet(1, mintTotalSamples)
        FileGet(1, mintLoopOffset)
        FileGet(1, mintLoopSamples)
        FileGet(1, mintRate)
        If Not (mbytVersion(2) = 1 And mbytVersion(3) = 1) Then mintRate = 60
        FileClose(1)
    End Sub

    Private Sub GetGD3Data(ByRef strFileName As String)
        'Verify file; if file is zipped, file will be unzipped
        Dim strFileExt As String = LCase(Path.GetExtension(strFileName))
        If (strFileExt = ".vgz") And (mstrUnzippedFileName = "") Then
            UnzipFile(strFileName)
        End If

        'Verify "Vgm "
        Dim strVgm_ As String, i As Integer
        If Not (mstrVgm_ = "Vgm ") Then
            Exit Sub
        End If
        If mintGD3Offset = 0 Then
            Exit Sub
        End If

        'Open file, retrieve GD3 tag, close file
        FileOpen(1, strFileName, OpenMode.Binary)
        Seek(1, (mintGD3Offset + 21))
        FileGet(1, mbytGd3_)
        Dim strGd3_ As String
        For i = 0 To 3
            strGd3_ &= Chr(mbytGd3_(i))
        Next
        If Not (strGd3_ = "Gd3 ") Then
            Exit Sub
        End If
        mstrGd3_ = strGd3_
        FileGet(1, mintGD3Version)
        FileGet(1, mintGD3Length)
        ReDim mshtGD3Data((mintGD3Length \ 2) - 1)
        FileGet(1, mshtGD3Data)
        FileClose(1)

        'Decode GD3 data
        Dim chrChar((mintGD3Length \ 2) - 1) As Char
        For i = 0 To UBound(chrChar)
            chrChar(i) = Chr(mshtGD3Data(i))
            If Not chrChar(i) = Chr(0) Then
                mstrTrackNameEnglish &= chrChar(i)
            Else
                Exit For
            End If
        Next
        For i = (i + 1) To UBound(chrChar)
            chrChar(i) = Chr(mshtGD3Data(i))
            If Not chrChar(i) = Chr(0) Then
                mstrTrackNameJapanese &= chrChar(i)
            Else
                Exit For
            End If
        Next
        For i = (i + 1) To UBound(chrChar)
            chrChar(i) = Chr(mshtGD3Data(i))
            If Not chrChar(i) = Chr(0) Then
                mstrGameNameEnglish &= chrChar(i)
            Else
                Exit For
            End If
        Next
        For i = (i + 1) To UBound(chrChar)
            chrChar(i) = Chr(mshtGD3Data(i))
            If Not chrChar(i) = Chr(0) Then
                mstrGameNameJapanese &= chrChar(i)
            Else
                Exit For
            End If
        Next
        For i = (i + 1) To UBound(chrChar)
            chrChar(i) = Chr(mshtGD3Data(i))
            If Not chrChar(i) = Chr(0) Then
                mstrSystemNameEnglish &= chrChar(i)
            Else
                Exit For
            End If
        Next
        For i = (i + 1) To UBound(chrChar)
            chrChar(i) = Chr(mshtGD3Data(i))
            If Not chrChar(i) = Chr(0) Then
                mstrSystemNameJapanese &= chrChar(i)
            Else
                Exit For
            End If
        Next
        For i = (i + 1) To UBound(chrChar)
            chrChar(i) = Chr(mshtGD3Data(i))
            If Not chrChar(i) = Chr(0) Then
                mstrOriginalAuthorEnglish &= chrChar(i)
            Else
                Exit For
            End If
        Next
        For i = (i + 1) To UBound(chrChar)
            chrChar(i) = Chr(mshtGD3Data(i))
            If Not chrChar(i) = Chr(0) Then
                mstrOriginalAuthorJapanese &= chrChar(i)
            Else
                Exit For
            End If
        Next
        For i = (i + 1) To UBound(chrChar)
            chrChar(i) = Chr(mshtGD3Data(i))
            If Not chrChar(i) = Chr(0) Then
                mstrGameReleaseDate &= chrChar(i)
            Else
                Exit For
            End If
        Next
        For i = (i + 1) To UBound(chrChar)
            chrChar(i) = Chr(mshtGD3Data(i))
            If Not chrChar(i) = Chr(0) Then
                mstrVGMAuthor &= chrChar(i)
            Else
                Exit For
            End If
        Next
        For i = (i + 1) To UBound(chrChar)
            chrChar(i) = Chr(mshtGD3Data(i))
            If Not chrChar(i) = Chr(0) Then
                mstrNotes &= chrChar(i)
            Else
                Exit For
            End If
        Next
    End Sub

    Private Sub GetVGMData(ByRef strFileName As String)
        'Verify file; if file is zipped, file will be unzipped
        Dim strFileExt As String = LCase(Path.GetExtension(strFileName))
        If (strFileExt = ".vgz") And (mstrUnzippedFileName = "") Then
            UnzipFile(strFileName)
        End If

        'Verify "Vgm "
        Dim strVgm_ As String, i As Integer
        If Not (mstrVgm_ = "Vgm ") Then
            Exit Sub
        End If

        'Open file and get VGM data
        FileOpen(1, strFileName, OpenMode.Binary)
        If mintGD3Offset = 0 Then
            ReDim mbytVGMData(LOF(1) - 65)
        Else
            ReDim mbytVGMData(mintGD3Offset - 41)
        End If
        Seek(1, 65)
        FileGet(1, mbytVGMData)
        FileClose(1)
    End Sub

    Public Sub ConvertVGMtoMID()
        Dim intFilePos As Integer
        Dim bytCommand As Byte, bytPort As Byte, bytRegister As Byte, bytData As Byte
        Dim sngWait As Single

        ' VGM Command Constants
        Const GG_STEREO = &H4F
        Const PSG = &H50
        Const YM2413 = &H51
        Const YM2612_P0 = &H52
        Const YM2612_P1 = &H53
        Const YM2151 = &H54
        Const WAIT_N_SAMPLES = &H61
        Const WAIT_735_SAMPLES = &H62
        Const WAIT_882_SAMPLES = &H63
        Const END_OF_SOUND_DATA = &H66

        ' Select PSG Constants
        Const PSG_CHANNEL_SELECT = &HF0 '240
        Const PSG_TONE_1 = &H80 '128
        Const PSG_TONE_2 = &HA0 '160
        Const PSG_TONE_3 = &HC0 '192
        Const PSG_NOISE = &HE0 '224
        Const PSG_ATTENUATOR = &H70 '112

        Dim MIDI As New clsMID()

        Dim pPSG As New clsSN76489() : pPSG.Init(MIDI)
        Dim yYM2413 As New clsYM2413() : yYM2413.Init(MIDI)
        Dim yYM2612 As New clsYM2612() : yYM2612.Init(MIDI)

        Do
            bytCommand = mbytVGMData(intFilePos)
            Select Case bytCommand
                Case GG_STEREO
                    bytData = mbytVGMData(intFilePos + 1)
                    intFilePos = intFilePos + 2
                Case PSG
                    bytData = mbytVGMData(intFilePos + 1)
                    Select Case (bytData And PSG_CHANNEL_SELECT)
                        Case PSG_TONE_1, PSG_TONE_2, PSG_TONE_3
                            pPSG.ProcessCommand(bytData, mbytVGMData(intFilePos + 3), MIDI)
                            intFilePos = intFilePos + 4
                        Case Else
                            pPSG.ProcessCommand(mbytVGMData(intFilePos + 1), 0, MIDI)
                            intFilePos = intFilePos + 2
                    End Select
                Case YM2413
                    bytRegister = mbytVGMData(intFilePos + 1)
                    bytData = mbytVGMData(intFilePos + 2)
                    yYM2413.ProcessCommand(bytRegister, bytData, MIDI)
                    intFilePos = intFilePos + 3
                Case YM2612_P0
                    bytPort = 0
                    bytRegister = mbytVGMData(intFilePos + 1)
                    bytData = mbytVGMData(intFilePos + 2)
                    yYM2612.ProcessCommand(bytPort, bytRegister, bytData, MIDI)
                    intFilePos = intFilePos + 3
                Case YM2612_P1
                    bytPort = 3
                    bytRegister = mbytVGMData(intFilePos + 1)
                    bytData = mbytVGMData(intFilePos + 2)
                    yYM2612.ProcessCommand(bytPort, bytRegister, bytData, MIDI)
                    intFilePos = intFilePos + 3
                Case YM2151
                    bytRegister = mbytVGMData(intFilePos + 1)
                    bytData = mbytVGMData(intFilePos + 2)
                    intFilePos = intFilePos + 3
                Case WAIT_N_SAMPLES
                    sngWait = mbytVGMData(intFilePos + 2)
                    sngWait *= 256
                    sngWait += mbytVGMData(intFilePos + 1)
                    sngWait /= 735
                    MIDI.msngDeltaTime += sngWait
                    intFilePos = intFilePos + 3
                Case WAIT_735_SAMPLES
                    sngWait = 1
                    MIDI.msngDeltaTime += sngWait
                    intFilePos = intFilePos + 1
                Case WAIT_882_SAMPLES
                    sngWait = 1
                    MIDI.msngDeltaTime += sngWait
                    intFilePos = intFilePos + 1
                Case END_OF_SOUND_DATA
                    Exit Do
                Case Else
                    intFilePos = intFilePos + 1
            End Select

            mstrConversionStatus = _
                "Converting VGM to MID... (" & CStr((intFilePos / UBound(mbytVGMData)) * 100) & ")"

            Application.DoEvents()

        Loop Until intFilePos >= UBound(mbytVGMData)

        MIDI.SaveFile(Path.GetFileNameWithoutExtension(mstrFileName) & ".mid", mintRate)

        MIDI = Nothing : pPSG = Nothing : yYM2413 = Nothing : yYM2612 = Nothing

    End Sub

    Private Sub GetGYMData(ByVal strFileName As String)
        'Open file and get GYM data
        FileOpen(1, strFileName, OpenMode.Binary)
        ReDim mbytGYMData(LOF(1) - 1)
        FileGet(1, mbytGYMData)
        FileClose(1)

        mintOffset = 0
        mbytVersion(0) = 0
        mbytVersion(1) = 0
        mbytVersion(2) = 1
        mbytVersion(3) = 1
        mintHzPSG = 3595740
        mintHzFM = 3595740
        mintGD3Offset = 0
        mintTotalSamples = 0
        mintLoopOffset = 0
        mintLoopSamples = 0
        mintRate = 60
    End Sub

    Private Sub ConvertGYMtoVGM()
        Dim lngGYMFilePos, lngVGMFilePos As Long
        Dim bytSwitch As Byte
        Dim bytPort, bytRegister, bytData, bytPSGMSB, bytPSGLSB As Byte

        ' GYM Command Constants
        Const DELAY_166ms = 0
        Const YM2612_PORT0 = 1
        Const YM2612_PORT1 = 2
        Const YM2612_PSG = 3

        ' VGM Command Constants
        Const GG_STEREO = &H4F
        Const PSG = &H50
        Const YM2413 = &H51
        Const YM2612_P0 = &H52
        Const YM2612_P1 = &H53
        Const YM2151 = &H54
        Const WAIT_N_SAMPLES = &H61
        Const WAIT_735_SAMPLES = &H62
        Const WAIT_882_SAMPLES = &H63
        Const END_OF_SOUND_DATA = &H66

        ReDim mbytVGMData(32767)
        Dim i As Integer

        Do
            bytSwitch = mbytGYMData(lngGYMFilePos)
            Select Case bytSwitch
                Case DELAY_166ms
                    i = mbytVGMData.GetUpperBound(0)
                    If lngVGMFilePos > i Then
                        ReDim Preserve mbytVGMData(i + 32767)
                    End If
                    mbytVGMData(lngVGMFilePos) = WAIT_735_SAMPLES
                    lngGYMFilePos += 1
                    lngVGMFilePos += 1
                Case YM2612_PORT0, YM2612_PORT1
                    bytPort = IIf(bytSwitch = YM2612_PORT0, YM2612_P0, YM2612_P1)
                    bytRegister = mbytGYMData(lngGYMFilePos + 1)
                    bytData = mbytGYMData(lngGYMFilePos + 2)
                    lngGYMFilePos += 3
                    i = mbytVGMData.GetUpperBound(0)
                    If (lngVGMFilePos + 2) > i Then
                        ReDim Preserve mbytVGMData(i + 32767)
                    End If
                    mbytVGMData(lngVGMFilePos) = bytPort
                    mbytVGMData(lngVGMFilePos + 1) = bytRegister
                    mbytVGMData(lngVGMFilePos + 2) = bytData
                    lngVGMFilePos += 3
                Case 3
                    bytPort = PSG
                    bytPSGMSB = mbytGYMData(lngGYMFilePos + 1)
                    lngGYMFilePos = lngGYMFilePos + 2
                    i = mbytVGMData.GetUpperBound(0)
                    If (lngVGMFilePos + 1) > i Then
                        ReDim Preserve mbytVGMData(i + 32767)
                    End If
                    mbytVGMData(lngVGMFilePos) = bytPort
                    mbytVGMData(lngVGMFilePos + 1) = bytPSGMSB
                    lngVGMFilePos = lngVGMFilePos + 2
                Case Else
                    lngGYMFilePos = lngGYMFilePos + 1
            End Select

            mstrConversionStatus = _
                "Converting GYM to VGM... (" & CStr((lngGYMFilePos / UBound(mbytGYMData)) * 100) & ")"

            Application.DoEvents()

        Loop Until lngGYMFilePos >= UBound(mbytGYMData)

        ReDim Preserve mbytVGMData(lngVGMFilePos)
        mbytGYMData = Nothing
    End Sub

End Class