Imports vgm2mid.clsMID
Public Class clsYM2413
    ' Base oscillator frequency
    Const OSC1 = 53693100
    ' Sample frequency (fsam)
    Const fM = OSC1 / 15
    Const fsam = fM / 72

    ' YM2413 Register Constants
    Enum Registers
        User = &H0
        Rhythm = &HE
        TEST = &HF
        FNumLSB = &H10
        SusKeyBlockFNumMSB = &H20
        InstVol = &H30
    End Enum

    Enum RegValues
        RhythmMode = 32
        BD = 16
        SD = 8
        TOM = 4
        TCT = 2
        HH = 1
        Sus = 32
        Key = 16
        Block = 14
        FNumMSB = 1
        Inst = 240
        Vol = 15
    End Enum

    ' Tone Data (INST) Constants
    Enum Instruments
        Original = 0
        Violin = 1
        Guitar = 2
        Piano = 3
        Flute = 4
        Clarinet = 5
        Oboe = 6
        Trumpet = 7
        Organ = 8
        Horn = 9
        Synthesizer = 10
        Harpsichord = 11
        Vibraphone = 12
        SynthBass = 13
        AcousticBass = 14
        ElectricGuitar = 15
    End Enum

    Public Sub Init(ByRef MIDI As clsMID)
        Dim ProgramChange As New clsMID.Commands.ProgramChange()
        Dim i As Integer
        For i = 0 To 8
            With ProgramChange
                .Channel = i : .Program = Programs.Lead1Square
            End With
            MIDI.AddEvent(ProgramChange)
        Next
    End Sub

    Private Function ToHz(ByVal FNum As Double, Optional ByVal Block As Byte = 0) As Double
        Return (FNum * fsam * 2 ^ (Block - 19))
    End Function

    Public Sub ProcessCommand(ByVal Register As Byte, ByVal Data As Byte, ByRef MIDI As clsMID)
        Static CH As Byte
        Static FNum_MSB(8) As Byte
        Static SUS_ON(8) As Byte, KEY_ON(8) As Byte, Block(8) As Byte, FNum_LSB(8) As Byte, FNum_1(8) As Long, FNum_2(8) As Long, Hz_1(8) As Double, Hz_2(8) As Double, Note_1(8) As Double, Note_2(8) As Double
        Static Instrument(8) As Byte, Volume(8) As Byte, BD_VOL As Byte, HH_VOL As Byte, SD_VOL As Byte, TOM_VOL As Byte, TCT_VOL As Byte
        Static MIDINote(8) As Byte, MIDIWheel(8) As Short, MIDIVolume(8) As Byte, NoteOn_1(8) As Byte, NoteOn_2(8) As Byte, Percussion_On(5) As Byte

        Static NoteOff As New Commands.NoteOff()
        Static NoteOn As New Commands.NoteOn()
        Static ControllerChange As New Commands.ControllerChange()
        Static ProgramChange As New Commands.ProgramChange()

        Select Case Register
            '    Case YM2413_REG_USER To (YM2413_REG_USER + 7)

        Case Registers.Rhythm
                CH = 9 'Midi channel 10
                If (Data And RegValues.BD) = 0 Then
                    With NoteOff
                        .Channel = CH : .Note = Percussion.AcousticBassDrum : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                Else
                    With NoteOff
                        .Channel = CH : .Note = Percussion.AcousticBassDrum : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                    With NoteOn
                        .Channel = CH : .Note = Percussion.AcousticBassDrum : .Velocity = BD_VOL
                    End With
                    MIDI.AddEvent(NoteOn)
                End If
                If (Data And RegValues.SD) = 0 Then
                    With NoteOff
                        .Channel = CH : .Note = Percussion.AcousticSnare : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                Else
                    With NoteOff
                        .Channel = CH : .Note = Percussion.AcousticSnare : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                    With NoteOn
                        .Channel = CH : .Note = Percussion.AcousticSnare : .Velocity = BD_VOL
                    End With
                    MIDI.AddEvent(NoteOn)
                End If
                If (Data And RegValues.TOM) = 0 Then
                    With NoteOff
                        .Channel = CH : .Note = Percussion.LowFloorTom : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                Else
                    With NoteOff
                        .Channel = CH : .Note = Percussion.LowFloorTom : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                    With NoteOn
                        .Channel = CH : .Note = Percussion.LowFloorTom : .Velocity = BD_VOL
                    End With
                    MIDI.AddEvent(NoteOn)
                End If
                If (Data And RegValues.TCT) = 0 Then
                    With NoteOff
                        .Channel = CH : .Note = Percussion.OpenHiHat : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                Else
                    With NoteOff
                        .Channel = CH : .Note = Percussion.OpenHiHat : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                    With NoteOn
                        .Channel = CH : .Note = Percussion.OpenHiHat : .Velocity = BD_VOL
                    End With
                    MIDI.AddEvent(NoteOn)
                End If
                If (Data And RegValues.HH) = 0 Then
                    With NoteOff
                        .Channel = CH : .Note = Percussion.ClosedHiHat : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                Else
                    With NoteOff
                        .Channel = CH : .Note = Percussion.ClosedHiHat : .Velocity = 0
                    End With
                    MIDI.AddEvent(NoteOff)
                    With NoteOn
                        .Channel = CH : .Note = Percussion.ClosedHiHat : .Velocity = BD_VOL
                    End With
                    MIDI.AddEvent(NoteOn)
                End If

                '    Case YM2413_REG_TEST

            Case Registers.FNumLSB To (Registers.FNumLSB + 8)
                CH = Register - Registers.FNumLSB
                'If YM2413_CH_DISABLED(CH) = 1 Then Exit Sub
                FNum_LSB(CH) = Data

            Case Registers.SusKeyBlockFNumMSB To (Registers.SusKeyBlockFNumMSB + 8)
                CH = Register - Registers.SusKeyBlockFNumMSB

                'If YM2413_CH_DISABLED(CH) = 1 Then Exit Sub

                SUS_ON(CH) = IIf((Data And RegValues.Sus) = 0, 0, 1)
                KEY_ON(CH) = IIf((Data And RegValues.Key) = 0, 0, 1)
                Block(CH) = ((Data And RegValues.Block) / 2)
                FNum_MSB(CH) = (Data And RegValues.FNumMSB)

                FNum_1(CH) = FNum_2(CH)
                FNum_2(CH) = (FNum_MSB(CH) * 256) + FNum_LSB(CH)

                Hz_1(CH) = Hz_2(CH)
                Hz_2(CH) = ToHz(FNum_2(CH), Block(CH))

                Note_1(CH) = Note_2(CH)
                Note_2(CH) = MIDI.toNote(Hz_2(CH))

                NoteOn_1(CH) = NoteOn_2(CH)
                NoteOn_2(CH) = KEY_ON(CH)

                If KEY_ON(CH) = 0 Then
                    If NoteOn_1(CH) <> 0 Then
                        With NoteOff
                            .Channel = CH : .Note = MIDINote(CH) : .Velocity = 0
                        End With
                        MIDI.AddEvent(NoteOff)
                    End If
                    NoteOn_2(CH) = 0
                ElseIf KEY_ON(CH) = 1 Then
                    If (NoteOn_2(CH) = 1 And NoteOn_1(CH) = 0) Then
                        MIDI.ChangePitch(Note_1(CH), Note_2(CH), CH, _
                            MIDINote(CH), MIDIWheel(CH), True)
                    Else
                        If Not (Note_2(CH) = Note_1(CH)) Then
                            MIDI.ChangePitch(Note_1(CH), Note_2(CH), CH, _
                                MIDINote(CH), MIDIWheel(CH), False)
                            NoteOn_2(CH) = 1
                        End If
                    End If
                End If

            Case Registers.InstVol To (Registers.InstVol + 5)
                CH = Register - Registers.InstVol

                'If YM2413_PROG_DISABLED(CH) = 1 Then Exit Sub

                Instrument(CH) = ((Data And RegValues.Inst) / 16)
                Volume(CH) = 127 - ((Data And RegValues.Vol) * 8)

                Select Case Instrument(CH)
                    Case Instruments.Original
                        With ProgramChange
                            .Channel = CH : .Program = Programs.Lead8BassLead
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Violin
                        With ProgramChange
                            .Channel = CH : .Program = Programs.StringEnsemble1
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Guitar
                        With ProgramChange
                            .Channel = CH : .Program = Programs.AcousticGuitarSteel
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Piano
                        With ProgramChange
                            .Channel = CH : .Program = Programs.AcousticGrandPiano
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Flute
                        With ProgramChange
                            .Channel = CH : .Program = Programs.Flute
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Clarinet
                        With ProgramChange
                            .Channel = CH : .Program = Programs.Clarinet
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Oboe
                        With ProgramChange
                            .Channel = CH : .Program = Programs.Oboe
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Trumpet
                        With ProgramChange
                            .Channel = CH : .Program = Programs.Trumpet
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Organ
                        With ProgramChange
                            .Channel = CH : .Program = Programs.HammondOrgan
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Horn
                        With ProgramChange
                            .Channel = CH : .Program = Programs.Trombone
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Synthesizer
                        With ProgramChange
                            .Channel = CH : .Program = Programs.Pad3Polysynth
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Harpsichord
                        With ProgramChange
                            .Channel = CH : .Program = Programs.Harpsichord
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.Vibraphone
                        With ProgramChange
                            .Channel = CH : .Program = Programs.Vibraphone
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.SynthBass
                        With ProgramChange
                            .Channel = CH : .Program = Programs.SynthBass1
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.AcousticBass
                        With ProgramChange
                            .Channel = CH : .Program = Programs.AcousticBass
                        End With
                        MIDI.AddEvent(ProgramChange)
                    Case Instruments.ElectricGuitar
                        With ProgramChange
                            .Channel = CH : .Program = Programs.ElectricGuitarJazz
                        End With
                        MIDI.AddEvent(ProgramChange)
                End Select

                'If YM2413_VOL_DISABLED(CH) = 1 Then Exit Sub

                With ControllerChange
                    .Controller = Controllers.Volume
                    .Channel = CH
                    .ControllerValue = Volume(CH)
                End With

                MIDI.AddEvent(ControllerChange)

            Case (Registers.InstVol + 6) To (Registers.InstVol + 8)
                'If YM2413_PERCUSSION_DISABLED = 1 Then Exit Sub

                Select Case Register
                    Case (Registers.InstVol + 6)
                        BD_VOL = (127 - ((Data And 15) * 8)) * 0.8
                    Case (Registers.InstVol + 7)
                        HH_VOL = 127 - (((Data And 240) / 16) * 8)
                        SD_VOL = (127 - ((Data And 15) * 8)) * 0.8
                    Case (Registers.InstVol + 8)
                        TOM_VOL = (127 - (((Data And 240) / 16) * 8)) * 0.6
                        TCT_VOL = 127 - ((Data And 15) * 8)
                End Select

            Case Else

        End Select

    End Sub

End Class