Imports vgm2mid.clsMID
Class clsYM2612

    ' Base oscillator frequency
    Const OSC1 = 53693100

    ' Sample frequency (fsam)
    Const fM = OSC1 / 14
    Const fsam = fM / 72

    ' YM2612 register select constants
    Const YM2612_TEST = &H21  ' LSI test data
    Const YM2612_LFO_EN_LFO_FREQ = &H22
    Const YM2612_LFO_EN = 8
    Const YM2612_LFO_FREQ = 7
    Const YM2612_TIMER_A_MSB = &H24
    Const YM2612_TIMER_A_LSB = &H25
    Const YM2612_TIMER_B = &H26
    Const YM2612_CH3_MODE_RESET_ENABLE_LOAD = &H27
    Const YM2612_CH3_MODE = 192
    Const YM2612_TIMER_RESET = 48
    Const YM2612_TIMER_ENABLE = 12
    Const YM2612_TIMER_LOAD = 3
    Const YM2612_SLOT_CH = &H28  ' Key On/Off
    Const YM2612_SLOT = 240
    Const YM2612_CH = 7
    Const YM2612_DAC = &H2A  ' DAC Data
    Const YM2612_DAC_DATA = 255
    Const YM2612_DAC_EN = &H2B  ' DAC Data
    Const YM2612_DAC_EN_BIT = 127
    Const YM2612_DT_MULTI = &H30  ' Detune/Multiple
    Const YM2612_DT = 112
    Const YM2612_MULTI = 15
    Const YM2612_TL = &H40  ' Total Level
    Const YM2612_KS_AR = &H50  ' Key Scale/Attack Rate
    Const YM2612_KS = 192
    Const YM2612_AR = 31
    Const YM2612_DR = &H60  ' Decay Rate
    Const YM2612_SR = &H70  ' Sustain Rate
    Const YM2612_SL_RR = &H80  ' Sustain Level/Release Rate
    Const YM2612_SL = 240
    Const YM2612_RR = 15
    Const YM2612_SSG_EG = &H90  ' SSG-Type Envelope Control
    Const YM2612_FNum_LSB = &HA0
    Const YM2612_Block_FNum_MSB = &HA4
    Const YM2612_Block = 56
    Const YM2612_FNum_MSB = 7
    Const YM2612_CH3_SPECIAL_FNum_MSB = &HA8
    Const YM2612_CH3_SPECIAL_BLOCK_FNum_LSB = &HAC
    Const YM2612_CH3_SPECIAL_Block = 56
    Const YM2612_CH3_SPECIAL_FNum_LSB = 7
    Const YM2612_FB_CONNECTION = &HB0  ' Self-Feedback/Connection
    Const YM2612_FB = 56
    Const YM2612_CONNECTION = 7
    Const YM2612_L_R = &HB4  ' Pan Select
    Const YM2612_L = 128
    Const YM2612_R = 64
    Const YM2612_BOTH_L_R = 192

    Public Sub Init(ByRef MIDI As clsMID)
        Dim ProgramChange As New clsMID.Commands.ProgramChange()
        Dim i As Integer
        For i = 0 To 8
            With ProgramChange
                .Channel = i : .Program = Programs.Lead8BassLead
            End With
            MIDI.AddEvent(ProgramChange)
        Next
    End Sub

    Function toHz(ByVal FNum As Double, Optional ByVal Block As Byte = 0) As Double
        Return (FNum * fsam * (2 ^ (Block - 21)))
    End Function

    Sub ProcessCommand(ByVal Port As Byte, ByVal Register As Byte, _
       ByVal Data As Byte, ByRef MIDI As clsMID)

        '    Static LFO_Freq As Byte
        '    Static TimerA_MSB As Byte, TimerA_LSB As Byte, TimerA As Integer
        '    Static TimerB As Byte
        '    Static CH3_Mode As Byte ', Timer_RESET As Byte, Timer_ENABLE As Byte, Timer_LOAD As Byte
        Static Slot(5) As Byte, CH As Byte, OP As Byte
        Static DAC_EN As Byte
        '    Static DT(5, 3) As Byte, MULTI(5, 3) As Byte
        Static TL(5, 3) As Byte
        '    Static KS(5, 3) As Byte, AR(5, 3) As Byte
        '    Static DR(5, 3) As Byte
        '    Static SR(5, 3) As Byte
        '    Static SL(5, 3) As Byte, RR(5, 3) As Byte
        '    Static SSG_EG(5, 3) As Byte
        Static FNum_MSB(5) As Byte
        Static Block(5) As Byte, FNum_LSB(5) As Byte, FNum_1(5) As Long, FNum_2(5) As Long, Hz_1(5) As Double, Hz_2(5) As Double, Note_1(5) As Double, Note_2(5) As Double
        '    Static CH3_Special_FNum_MSB(2) As Byte
        '    Static CH3_Special_Block(2) As Byte, CH3_Special_FNum_LSB(2) As Byte, CH3_Special_FNum_1(2) As Integer, CH3_Special_FNum_2(2) As Integer
        Static Feedback(5) As Byte, Connection(5) As Byte
        Static MIDINote(5) As Byte, MIDIWheel(5) As Integer, MIDIVolume(5) As Byte, MIDIPan(5) As Byte, NoteOn_1(5) As Byte, NoteOn_2(5) As Byte

        Static NoteOff As New Commands.NoteOff()
        Static NoteOn As New Commands.NoteOn()
        Static ControllerChange As New Commands.ControllerChange()
        Static ProgramChange As New Commands.ProgramChange()

        Select Case Register
            '    Case YM2612_TEST

            '    Case YM2612_LFO_EN_LFO_FREQ

            '    Case YM2612_TIMER_A_MSB

            '    Case YM2612_TIMER_A_LSB

            '    Case YM2612_TIMER_B

            '    Case YM2612_CH3_MODE_RESET_ENABLE_LOAD
            '        CH3_Mode = (Data And YM2612_CH3_MODE) / 96

        Case YM2612_SLOT_CH
                CH = (Data And YM2612_CH)
                CH = IIf(CH > 3, CH - 1, CH)

                CH = IIf(CH > 5, 5, CH)

                'If YM2612_CH_DISABLED(CH) = 1 Then Exit Sub

                Slot(CH) = (Data And YM2612_SLOT) / 8

                If Slot(CH) = 0 Then
                    If NoteOn_2(CH) <> 0 Then
                        MIDIWheel(CH) = ControllerValues.PitchWheelCenter
                        With NoteOff
                            .Channel = CH : .Note = MIDINote(CH) : .Velocity = 0
                        End With
                        MIDI.AddEvent(NoteOff)
                        NoteOn_1(CH) = NoteOn_2(CH)
                        NoteOn_2(CH) = 0
                    End If
                Else
                    If NoteOn_2(CH) = 0 Then
                        'MIDI.ChangePitch(Note_1(CH), Note_2(CH), CH, MIDINote(CH), MIDIWheel(CH), True)
                        MIDI.NewNote(Note_1(CH), Note_2(CH), CH, MIDINote(CH), MIDIWheel(CH))
                        NoteOn_1(CH) = NoteOn_2(CH)
                        NoteOn_2(CH) = 1
                    End If
                End If

                '    Case YM2612_DAC

            Case YM2612_DAC_EN
                If (Register And YM2612_DAC_EN) = YM2612_DAC_EN Then DAC_EN = 1 Else DAC_EN = 0

                '    Case YM2612_DT_MULTI

            Case YM2612_TL To (YM2612_TL + &HF)
                CH = ((Register Xor YM2612_TL) Mod 4) + Port

                'If YM2612_CH_DISABLED(CH) = 1 Then Exit Sub
                'If YM2612_VOL_DISABLED(CH) = 1 Then Exit Sub

                OP = Fix((Register Xor YM2612_TL) / 4)

                CH = IIf(CH > 5, 5, CH)
                TL(CH, OP) = (Data And &H7F)

                Select Case Connection(CH)
                    Case 0, 1, 2, 3
                        MIDIVolume(CH) = TL(CH, 3)
                    Case 4
                        MIDIVolume(CH) = ((TL(CH, 1) \ 2) + (TL(CH, 3) \ 2))
                    Case 5, 6
                        MIDIVolume(CH) = ((TL(CH, 1) \ 3) + (TL(CH, 2) \ 3) + (TL(CH, 3) \ 3))
                    Case 7
                        MIDIVolume(CH) = ((TL(CH, 0) \ 4) + (TL(CH, 1) \ 4) + (TL(CH, 2) \ 4) + (TL(CH, 3) \ 4))
                End Select

                Const Phase = Math.PI / 254
                MIDIVolume(CH) = ControllerValues.VolumeMax * (1 - (Math.Sin(MIDIVolume(CH) * Phase)))

                With ControllerChange
                    .Channel = CH : .Controller = Controllers.Volume : .ControllerValue = MIDIVolume(CH)
                End With
                MIDI.AddEvent(ControllerChange)

                '    Case YM2612_KS_AR

                '    Case YM2612_DR

                '    Case YM2612_SR

                '    Case YM2612_SL_RR

                '    Case YM2612_SSG_EG

            Case &HA4 To &HA6
                CH = (Register Xor &HA4) + Port

                'If YM2612_CH_DISABLED(CH) = 1 Then Exit Sub

                Block(CH) = (Data And YM2612_Block) / 8
                FNum_MSB(CH) = (Data And 7)
            Case &HA0 To &HA2
                CH = (Register Xor &HA0) + Port

                'If YM2612_CH_DISABLED(CH) = 1 Then Exit Sub

                FNum_LSB(CH) = Data

                FNum_1(CH) = FNum_2(CH)
                FNum_2(CH) = ((FNum_MSB(CH) * 256) + FNum_LSB(CH))

                Hz_1(CH) = Hz_2(CH)
                Hz_2(CH) = toHz(FNum_2(CH), Block(CH))

                Note_1(CH) = Note_2(CH)
                Note_2(CH) = MIDI.toNote(Hz_2(CH))

                'If NoteOn_2(CH) = 1 Then
                If Note_2(CH) <> Note_1(CH) Then
                    MIDI.ChangePitch(Note_1(CH), Note_2(CH), CH, MIDINote(CH), MIDIWheel(CH))
                    'MIDI.NewNote(Note_1(CH), Note_2(CH), CH, MIDINote(CH), MIDIWheel(CH))
                End If
                'End If

                '    Case YM2612_CH3_SPECIAL_FNum_MSB To (YM2612_CH3_SPECIAL_FNum_MSB + 2)

                '    Case YM2612_CH3_SPECIAL_BLOCK_FNum_LSB To (YM2612_CH3_SPECIAL_BLOCK_FNum_LSB + 2)

            Case YM2612_FB_CONNECTION To (YM2612_FB_CONNECTION + 2)
                    CH = (Register Xor YM2612_FB_CONNECTION) + Port
                    Connection(CH) = (Data And YM2612_CONNECTION)

                    'If YM2612_PROG_DISABLED(CH) = True Then Exit Sub

                    With ProgramChange
                        .Channel = CH : .Program = Programs.Lead8BassLead
                    End With
                    MIDI.AddEvent(ProgramChange)

            Case YM2612_L_R To (YM2612_L_R + 2)
                    CH = (Register Xor YM2612_L_R) + Port

                    'If YM2612_CH_DISABLED(CH) = 1 Then Exit Sub
                    'If YM2612_PAN_DISABLED(CH) = 1 Then Exit Sub

                    Select Case (Data And YM2612_BOTH_L_R)
                        Case YM2612_L
                            If MIDIPan(CH) <> ControllerValues.PanLeft Then
                                MIDIPan(CH) = ControllerValues.PanLeft
                                With ControllerChange
                                    .Channel = CH : .Controller = Controllers.Pan
                                    .ControllerValue = ControllerValues.PanLeft
                                End With
                                MIDI.AddEvent(ControllerChange)
                            End If
                        Case YM2612_R
                            If MIDIPan(CH) <> ControllerValues.PanRight Then
                                MIDIPan(CH) = ControllerValues.PanRight
                                With ControllerChange
                                    .Channel = CH : .Controller = Controllers.Pan
                                    .ControllerValue = ControllerValues.PanRight
                                End With
                                MIDI.AddEvent(ControllerChange)
                            End If
                        Case YM2612_BOTH_L_R, 0
                            If MIDIPan(CH) <> ControllerValues.PanCenter Then
                                MIDIPan(CH) = ControllerValues.PanCenter
                                With ControllerChange
                                    .Channel = CH : .Controller = Controllers.Pan
                                    .ControllerValue = ControllerValues.PanCenter
                                End With
                                MIDI.AddEvent(ControllerChange)
                            End If
                    End Select

                    'If YM2612_DAC_DISABLED = 1 Then Exit Sub

                    If CH = 5 And DAC_EN = 1 Then
                        With ControllerChange
                            .Channel = 9 : .Controller = Controllers.Pan
                            .ControllerValue = MIDIPan(CH)
                        End With
                        MIDI.AddEvent(ControllerChange)
                        With NoteOff
                            .Channel = 9 : .Note = Percussion.OpenHiHat : .Velocity = 0
                        End With
                        MIDI.AddEvent(NoteOff)
                        With NoteOn
                            .Channel = 9 : .Note = Percussion.OpenHiHat : .Velocity = 95
                        End With
                        MIDI.AddEvent(NoteOn)
                    End If

                    '    Case Else

        End Select

    End Sub

End Class